

struct database {
};


#if SWIGVERSIONMINOR == 1 && SWIGVERSIONMAJOR == 1
%addmethods
#else
%extend
#endif
	database {
	database() {
		struct database* newdb = (struct database*)  Tcl_Alloc(sizeof(struct database));
		SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &newdb->env);
		SQLSetEnvAttr(newdb->env, SQL_ATTR_ODBC_VERSION,(void *) SQL_OV_ODBC3, 0);
		return newdb;
	}
	~database() {
		SQLFreeHandle(SQL_HANDLE_ENV, self->env);
		Tcl_Free((char *)self);
	}
	Tcl_Result drivers(Tcl_Interp *interp) {
	    char driver[256];
	    char attr[256];
	    char *p, *q, *r;
	    SQLSMALLINT driver_ret;
	    SQLSMALLINT attr_ret;
	    SQLUSMALLINT direction;
	    SQLRETURN ret;
	    Tcl_Obj  *oneDriver;
	    Tcl_Obj  *result = Tcl_NewObj();

	    direction = SQL_FETCH_FIRST;
	    while(SQL_SUCCEEDED(ret = SQLDrivers(self->env, direction,
						 driver, sizeof(driver), &driver_ret,
						 attr, sizeof(attr), &attr_ret))) {
		direction = SQL_FETCH_NEXT;
		oneDriver = Tcl_NewObj();
		if (Tcl_ListObjAppendElement(interp,oneDriver,Tcl_NewStringObj(driver, driver_ret)) != TCL_OK) return TCL_ERROR;
		r = &attr[attr_ret];
		q = attr;
		for (p = strchr(q,'\0'); p != NULL && p < r; p = strchr(q,'\0')) {
			q = p+1;
			*p = ';';
		}
		if (Tcl_ListObjAppendElement(interp,oneDriver,Tcl_NewStringObj(attr, attr_ret)) != TCL_OK) return TCL_ERROR;
		if (Tcl_ListObjAppendElement(interp,result,oneDriver) != TCL_OK) return TCL_ERROR;
	    }
	    Tcl_SetObjResult(interp,result);
	    return TCL_OK;
	}
	Tcl_Result datasources(Tcl_Interp *interp) {
	    char dsn[256];
	    char desc[256];
	    SQLSMALLINT dsn_ret;
	    SQLSMALLINT desc_ret;
	    SQLUSMALLINT direction;
	    SQLRETURN ret;
	    Tcl_Obj  *oneDatasource;
	    Tcl_Obj  *result = Tcl_NewObj();

	    direction = SQL_FETCH_FIRST;
	    while(SQL_SUCCEEDED(ret = SQLDataSources(self->env, direction,
					   dsn, sizeof(dsn), &dsn_ret,
					   desc, sizeof(desc), &desc_ret))) {
		direction = SQL_FETCH_NEXT;
		oneDatasource = Tcl_NewObj();
		if (Tcl_ListObjAppendElement(interp,oneDatasource,Tcl_NewStringObj(dsn, dsn_ret)) != TCL_OK) return TCL_ERROR;
		if (Tcl_ListObjAppendElement(interp,oneDatasource,Tcl_NewStringObj(desc, desc_ret)) != TCL_OK) return TCL_ERROR;
		if (Tcl_ListObjAppendElement(interp,result,oneDatasource) != TCL_OK) return TCL_ERROR;
	    }
	    Tcl_SetObjResult(interp,result);
	    return TCL_OK;
	}
	struct connection *connect (Tcl_Interp *interp,const char *connectionString,Tcl_Obj *opts=NULL,char **outmessage=NULL) {
		char numberbuffer[TCL_INTEGER_SPACE];
		Tcl_DString dstring;
		struct connection *newconnect;
		SQLRETURN ret;
		newconnect = (struct connection *) Tcl_Alloc(sizeof(struct connection));
		newconnect->parent = self;
		SQLAllocHandle(SQL_HANDLE_DBC,self->env,&newconnect->dbc);
		if (opts != NULL) {
		    int len,i;
		    Tcl_Obj *k, *v;
		    const char *ks;
		    if (Tcl_ListObjLength(interp,opts,&len) != TCL_OK) {
		    	if (outmessage != NULL) {
		    	    const char *r = Tcl_GetStringResult(interp);
			    *outmessage = Tcl_Alloc(strlen(r)+1);
			    strcpy(*outmessage,r);
			}
			SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
			Tcl_Free((char *)newconnect);
			return NULL;
		    }
		    if ((len & 1) == 1) {
		    	if (outmessage != NULL) {
			    sprintf(numberbuffer,"%d",len);
			    Tcl_DStringInit(&dstring);
			    Tcl_DStringAppend(&dstring,"Option List has an odd number of elements: ",-1);
			    Tcl_DStringAppend(&dstring,numberbuffer,-1);
			    Tcl_DStringAppend(&dstring,"!",1);
			    *outmessage = Tcl_Alloc(Tcl_DStringLength(&dstring)+1);
			    strcpy(*outmessage,Tcl_DStringValue(&dstring));
			    Tcl_DStringFree(&dstring);
		    	}
			SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
			Tcl_Free((char *)newconnect);
			return NULL;
		    }
		    for (i = 0; i < len; i += 2) {
		    	Tcl_ListObjIndex(interp,opts,i,&k);
		    	Tcl_ListObjIndex(interp,opts,i+1,&v);
		    	ks = Tcl_GetString(k);
		    	if (strncasecmp(ks,"-access_mode",strlen(ks)) == 0) {
		    	    const char *m = Tcl_GetString(v);
		    	    SQLUINTEGER mode = SQL_MODE_READ_WRITE;
		    	    if (strncasecmp(m,"read",strlen(m)) == 0) {
		    	    	mode = SQL_MODE_READ_ONLY;
		    	    } else if (strncasecmp(m,"write",strlen(m)) == 0) {
		    	    	mode = SQL_MODE_READ_WRITE;
		    	    } else {
		    	    	if (outmessage != NULL) {
				    Tcl_DStringInit(&dstring);
				    Tcl_DStringAppend(&dstring,"Undefined mode: \"",-1);
				    Tcl_DStringAppend(&dstring,m,-1);
				    Tcl_DStringAppend(&dstring,"\"!",-1);
				    *outmessage = Tcl_Alloc(Tcl_DStringLength(&dstring)+1);
				    strcpy(*outmessage,Tcl_DStringValue(&dstring));
				    Tcl_DStringFree(&dstring);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
		    	    }
			    ret = SQLSetConnectAttr(newconnect->dbc,SQL_ATTR_ACCESS_MODE,
		    	    			(SQLPOINTER)mode,0);
			    if (!SQL_SUCCEEDED(ret)) {
				if (outmessage != NULL) {
				    extract_error("connect: -access_mode",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
		    	} else if (strncasecmp(ks,"-current_catalog",strlen(ks)) == 0) {
		    	    const char *m = Tcl_GetString(v);
		    	    ret = SQLSetConnectAttr(newconnect->dbc,SQL_ATTR_CURRENT_CATALOG,
			    			(SQLPOINTER)m,SQL_NTS);
			    if (!SQL_SUCCEEDED(ret)) {
				if (outmessage != NULL) {
				    extract_error("connect: -current_catalog",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
		    	} else if (strncasecmp(ks,"-timeout",strlen(ks)) == 0) {
		    	    SQLUINTEGER timeout;
		    	    if (Tcl_GetLongFromObj(interp,v,&timeout) != TCL_OK) {
		    	    	if (outmessage != NULL) {
				    const char *r = Tcl_GetStringResult(interp);
				    *outmessage = Tcl_Alloc(strlen(r)+1);
				    strcpy(*outmessage,r);
			        }
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			    ret = SQLSetConnectAttr(newconnect->dbc,SQL_ATTR_LOGIN_TIMEOUT,
			    			(SQLPOINTER)timeout,0);
			    if (!SQL_SUCCEEDED(ret)) {
				if (outmessage != NULL) {
				    extract_error("connect: -timeout",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			} else if (strncasecmp(ks,"-odbc_cursors",strlen(ks)) == 0) {
			    SQLUINTEGER cursor = SQL_CUR_USE_DRIVER;
			    const char *m = Tcl_GetString(v);
			    if (strncasecmp(m,"if_needed",strlen(m)) == 0) {
			    	cursor = SQL_CUR_USE_IF_NEEDED;
			    } else if (strncasecmp(m,"odbc",strlen(m)) == 0) {
			    	cursor = SQL_CUR_USE_ODBC;
			    } else if (strncasecmp(m,"driver",strlen(m)) == 0) {
			    	cursor = SQL_CUR_USE_DRIVER;
			    } else {
		    	    	if (outmessage != NULL) {
				    Tcl_DStringInit(&dstring);
				    Tcl_DStringAppend(&dstring,"Undefined cursor setting: \"",-1);
				    Tcl_DStringAppend(&dstring,m,-1);
				    Tcl_DStringAppend(&dstring,"\"!",-1);
				    *outmessage = Tcl_Alloc(Tcl_DStringLength(&dstring)+1);
				    strcpy(*outmessage,Tcl_DStringValue(&dstring));
				    Tcl_DStringFree(&dstring);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			    ret = SQLSetConnectAttr(newconnect->dbc,SQL_ATTR_ODBC_CURSORS,
			    			(SQLPOINTER)cursor,0);
			    if (!SQL_SUCCEEDED(ret)) {
				if (outmessage != NULL) {
				    extract_error("connect: -odbc_cursors",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			} else if (strncasecmp(ks,"-packet_size",strlen(ks)) == 0) {
			    SQLUINTEGER packetsz;
		    	    if (Tcl_GetLongFromObj(interp,v,&packetsz) != TCL_OK) {
		    	    	if (outmessage != NULL) {
				    const char *r = Tcl_GetStringResult(interp);
				    *outmessage = Tcl_Alloc(strlen(r)+1);
				    strcpy(*outmessage,r);
			        }
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			    ret = SQLSetConnectAttr(newconnect->dbc,SQL_ATTR_PACKET_SIZE,
			    			(SQLPOINTER)packetsz,0);
			    if (!SQL_SUCCEEDED(ret)) {
				if (outmessage != NULL) {
				    extract_error("connect: -packet_size",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
				}
				SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
				Tcl_Free((char *)newconnect);
				return NULL;
			    }
			} else {
			    if (outmessage != NULL) {
				Tcl_DStringInit(&dstring);
				Tcl_DStringAppend(&dstring,"Undefined option keyword: \"",-1);
				Tcl_DStringAppend(&dstring,ks,-1);
				Tcl_DStringAppend(&dstring,"\", should be one of: ",-1);
				Tcl_DStringAppend(&dstring,"-access_mode ",-1);
				Tcl_DStringAppend(&dstring,"-current_catalog ",-1);
				Tcl_DStringAppend(&dstring,"-timeout ",-1);
				Tcl_DStringAppend(&dstring,"-odbc_cursors ",-1);
				Tcl_DStringAppend(&dstring,"-packet_size ",-1);
				*outmessage = Tcl_Alloc(Tcl_DStringLength(&dstring)+1);
				strcpy(*outmessage,Tcl_DStringValue(&dstring));
				Tcl_DStringFree(&dstring);
			    }
			    SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
			    Tcl_Free((char *)newconnect);
			    return NULL;
			}
		    }
		    	
		}
		ret = SQLDriverConnect(newconnect->dbc,NULL,
					connectionString,SQL_NTS,
					NULL,0,NULL,SQL_DRIVER_NOPROMPT);
		if (!SQL_SUCCEEDED(ret)) {
		    if (outmessage != NULL && ret != SQL_NO_DATA) {
			extract_error("connect: SQLDriverConnect",newconnect->dbc,SQL_HANDLE_DBC,outmessage);
		    } else if (outmessage != NULL && ret == SQL_NO_DATA) {
			const char *r = "connect: SQLDriverConnect: Missing connection string attributes!";
			*outmessage = Tcl_Alloc(strlen(r)+1);
			strcpy(*outmessage,r);
		    }
		    SQLFreeHandle(SQL_HANDLE_DBC,newconnect->dbc);
		    Tcl_Free((char *)newconnect);
		    return NULL;
		} else {
		    newconnect->attr_async_enable = SQL_ASYNC_ENABLE_OFF;
		    newconnect->attr_concurrency = SQL_CONCUR_READ_ONLY;
		    newconnect->attr_cursor_scrollable = SQL_NONSCROLLABLE;
		    newconnect->attr_cursor_sensitivity = SQL_UNSPECIFIED;
		    newconnect->attr_cursor_type = SQL_CURSOR_FORWARD_ONLY;
		    newconnect->attr_enable_auto_ipd = SQL_FALSE;
		    newconnect->attr_keyset_size = 0;
		    newconnect->attr_max_length = 0;
		    newconnect->attr_max_rows = 0;
		    newconnect->attr_metadata_id = SQL_FALSE;
		    newconnect->attr_noscan = SQL_NOSCAN_OFF;
		    newconnect->attr_query_timeout = 0;
		    newconnect->attr_simulate_cursor = 99; /* No appearent default */
		    newconnect->attr_use_bookmarks = SQL_UB_OFF;
		    newconnect->multiSetsP = 0;
		    return newconnect;
		}
		
	}
}


