%{

static SQLSMALLINT MapSqlType (SQLINTEGER colType) {
	switch (colType) {
	    case SQL_VARBINARY:
	    case SQL_LONGVARBINARY:
		return SQL_C_BINARY;
	    default:
		return SQL_C_CHAR;
	}
}

SQLSMALLINT LookUpSqlType(const char *typeName) {
	if (strcasecmp(typeName,"bigint") == 0) return SQL_BIGINT;
	else if (strcasecmp(typeName,"binary") == 0) return SQL_BINARY;
	else if (strcasecmp(typeName,"bit") == 0) return SQL_BIT;
	else if (strcasecmp(typeName,"char") == 0) return SQL_CHAR;
	else if (strcasecmp(typeName,"date") == 0) return SQL_DATE;
	else if (strcasecmp(typeName,"decimal") == 0) return SQL_DECIMAL;
	else if (strcasecmp(typeName,"double") == 0) return SQL_DOUBLE;
	else if (strcasecmp(typeName,"float") == 0) return SQL_FLOAT;
	else if (strcasecmp(typeName,"integer") == 0) return SQL_INTEGER;
	else if (strcasecmp(typeName,"longvarbinary") == 0) return SQL_LONGVARBINARY;
	else if (strcasecmp(typeName,"longvarchar") == 0) return SQL_LONGVARCHAR;
	else if (strcasecmp(typeName,"numeric") == 0) return SQL_NUMERIC;
	else if (strcasecmp(typeName,"real") == 0) return SQL_REAL;
	else if (strcasecmp(typeName,"smallint") == 0) return SQL_SMALLINT;
	else if (strcasecmp(typeName,"time") == 0) return SQL_TIME;
	else if (strcasecmp(typeName,"timestamp") == 0) return SQL_TIMESTAMP;
	else if (strcasecmp(typeName,"tinyint") == 0) return SQL_TINYINT;
	else if (strcasecmp(typeName,"varbinary") == 0) return SQL_VARBINARY;
	else if (strcasecmp(typeName,"varchar") == 0) return SQL_VARCHAR;
        else return SQL_UNKNOWN_TYPE;
}                                                               

static int NumResultCols(struct statement *statement,char **outmessage) {
	SQLRETURN ret;

	if (statement->numColumns < 0) {
		while ((ret = SQLNumResultCols(statement->stmt,
					       &statement->numColumns))
			== SQL_STILL_EXECUTING) EventWait(0);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("NumResultCols",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return -1;
		}
	}
	return statement->numColumns;
}	

static void EventWait(int delay) {
	if (!Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT) && delay) {
		Tcl_Sleep(delay);
	}
}

static struct statement *CreateStatement(Tcl_Interp *interp,
				  const struct connection *parent,
				  const char *sql,Tcl_Obj *argtypedefs,
				  char **outmessage) {
	struct statement *newstatement;
	SQLRETURN ret;
	int i, len;
	struct ArgDef *aD;
	newstatement = (struct statement *) Tcl_Alloc(sizeof(struct statement));
	memset(newstatement,0,sizeof(struct statement));
	newstatement->parent = parent;
	newstatement->numColumns = -1;
	SQLAllocHandle(SQL_HANDLE_STMT,parent->dbc,&newstatement->stmt);
	newstatement->multiSetsP = parent->multiSetsP;
#ifdef ERRORCHECK
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_ASYNC_ENABLE,
			     (SQLPOINTER)parent->attr_async_enable,
			     SQL_IS_UINTEGER);
	fprintf(stderr,"*** CreateStatement: ret = %d\n",ret);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_ASYNC_ENABLE)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#endif
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_CONCURRENCY,
			     (SQLPOINTER)parent->attr_concurrency,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_CONCURRENCY)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#ifdef ERRORCHECK
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_CURSOR_SCROLLABLE,
			     (SQLPOINTER)parent->attr_cursor_scrollable,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_CURSOR_SCROLLABLE)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#endif
#ifdef ERRORCHECK
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_CURSOR_SENSITIVITY,
			     (SQLPOINTER)parent->attr_cursor_sensitivity,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement  (SQL_ATTR_CURSOR_SENSITIVITY)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#endif
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_CURSOR_TYPE,
			     (SQLPOINTER)parent->attr_cursor_type,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_CURSOR_TYPE)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#ifdef ERRORCHECK
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_ENABLE_AUTO_IPD,
			     (SQLPOINTER)parent->attr_enable_auto_ipd,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_ENABLE_AUTO_IPD)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#endif
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_KEYSET_SIZE,
			     (SQLPOINTER)parent->attr_keyset_size,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_KEYSET_SIZE)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_MAX_LENGTH,
			     (SQLPOINTER)parent->attr_max_length,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_MAX_LENGTH)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_MAX_ROWS,
			     (SQLPOINTER)parent->attr_max_rows,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement  (SQL_ATTR_MAX_ROWS)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#ifdef ERRORCHECK
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_METADATA_ID,
			     (SQLPOINTER)parent->attr_metadata_id,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_METADATA_ID)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
#endif
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_NOSCAN,
			     (SQLPOINTER)parent->attr_noscan,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_NOSCAN)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_QUERY_TIMEOUT,
			     (SQLPOINTER)parent->attr_query_timeout,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_QUERY_TIMEOUT)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
	/* no default for attr_simulate_cursor */
	if (parent->attr_simulate_cursor != 99) {
		ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_SIMULATE_CURSOR,
				     (SQLPOINTER)parent->attr_simulate_cursor,
				     SQL_IS_UINTEGER);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("CreateStatement (SQL_ATTR_SIMULATE_CURSOR)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
			SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
			Tcl_Free((char *)newstatement);
			return NULL;
		}
	}
	ret = SQLSetStmtAttr(newstatement->stmt,SQL_ATTR_USE_BOOKMARKS,
			     (SQLPOINTER)parent->attr_use_bookmarks,
			     SQL_IS_UINTEGER);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("CreateStatement (SQL_ATTR_USE_BOOKMARKS)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
		SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
		Tcl_Free((char *)newstatement);
		return NULL;
	}
	if (strcasecmp(sql,"tables") == 0) {
		newstatement->statementType = STMT_TABLES;
	} else if (strcasecmp(sql,"columns") == 0) {
		newstatement->statementType = STMT_COLUMNS;
	} else if (strcasecmp(sql,"indexes") == 0) {
		newstatement->statementType = STMT_INDEXES;
	} else if (strcasecmp(sql,"typeinfo") == 0) {
		newstatement->statementType = STMT_TYPEINFO;
	} else if (strcasecmp(sql,"primarykeys") == 0) {
		newstatement->statementType = STMT_PRIMARYKEYS;
	} else {
		newstatement->statementType = STMT_SQL;
		while ((ret = SQLPrepare(newstatement->stmt,sql,strlen(sql)))
			    == SQL_STILL_EXECUTING) Tcl_Sleep(0);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("CreateStatement (SQLPrepare)",newstatement->stmt,SQL_HANDLE_STMT,outmessage);
			SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
			Tcl_Free((char *)newstatement);
			return NULL;
		}
		ret = SQLNumParams(newstatement->stmt,&(newstatement->numArgDefs));
		if (!SQL_SUCCEEDED(ret)) {
			newstatement->numArgDefs = 32;
		}
		if (newstatement->numArgDefs > 0)
			newstatement->argDefs =
				(struct ArgDef *) Tcl_Alloc(newstatement->numArgDefs*sizeof(struct ArgDef));
		aD = newstatement->argDefs;
		memset(aD,0,newstatement->numArgDefs*sizeof(struct ArgDef));
		for (i = 0; i < newstatement->numArgDefs; i++) {
			while((ret = SQLDescribeParam(newstatement->stmt, i,
						      &aD[i].sqlType,
						      &aD[i].columnDef,
						      &aD[i].scale,
						      &aD[i].nullableP)) == SQL_STILL_EXECUTING) Tcl_Sleep(0);
			if (!SQL_SUCCEEDED(ret)) {
				aD[i].sqlType = SQL_VARCHAR;
				aD[i].columnDef = 0;
				aD[i].scale = 0;
				aD[i].nullableP = SQL_NULLABLE_UNKNOWN;
			}
			aD[i].sourceType = MapSqlType(aD[i].sqlType);
		}
		if (argtypedefs != NULL) {
			if (Tcl_ListObjLength(interp,argtypedefs,&len) != TCL_OK) {
				if (outmessage != NULL) {
					const char *r = Tcl_GetStringResult(interp);
					*outmessage = Tcl_Alloc(strlen(r)+1);
					strcpy(*outmessage,r);
				}
				SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
				Tcl_Free((char *)newstatement);
				return NULL;
			}
			if (len > newstatement->numArgDefs) {
				/* Too Many arg defs? */
			}
			for (i = 0; i < len; i++) {
				Tcl_Obj *td,*tdele;
				int tdlen, tempint;
				if (Tcl_ListObjIndex(interp,argtypedefs,i,&td) != TCL_OK) {
					if (outmessage != NULL) {
						const char *r = Tcl_GetStringResult(interp);
						*outmessage = Tcl_Alloc(strlen(r)+1);
						strcpy(*outmessage,r);
					}
					SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
					Tcl_Free((char *) aD);
					Tcl_Free((char *)newstatement);
					return NULL;
				}
				if (Tcl_ListObjLength(interp,td,&tdlen) != TCL_OK) {
					if (outmessage != NULL) {
						const char *r = Tcl_GetStringResult(interp);
						*outmessage = Tcl_Alloc(strlen(r)+1);
						strcpy(*outmessage,r);
					}
					SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
					Tcl_Free((char *) aD);
					Tcl_Free((char *)newstatement);
					return NULL;
				}
				switch (tdlen) {
				    case 3:
					Tcl_ListObjIndex(interp,td,2,&tdele);
					if (Tcl_GetIntFromObj(interp,tdele,&tempint) != TCL_OK) {
						if (outmessage != NULL) {
							const char *r = Tcl_GetStringResult(interp);
							*outmessage = Tcl_Alloc(strlen(r)+1);
							strcpy(*outmessage,r);
						}
						SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
						Tcl_Free((char *) aD);
						Tcl_Free((char *)newstatement);
						return NULL;
					}
					aD[i].scale = tempint;
				    case 2:
					Tcl_ListObjIndex(interp,td,1,&tdele);
					if (Tcl_GetIntFromObj(interp,tdele,&tempint) != TCL_OK) {
						if (outmessage != NULL) {
							const char *r = Tcl_GetStringResult(interp);
							*outmessage = Tcl_Alloc(strlen(r)+1);
							strcpy(*outmessage,r);
						}
						SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
						Tcl_Free((char *) aD);
						Tcl_Free((char *)newstatement);
						return NULL;
					}
					aD[i].columnDef = tempint;
				    case 1:
					Tcl_ListObjIndex(interp,td,0,&tdele);
				    	aD[i].sqlType = LookUpSqlType(Tcl_GetString(tdele));
				    	aD[i].sourceType = MapSqlType(aD[i].sqlType);
				    	break;
				    default:
					if (outmessage != NULL) {
						const char *r = "Invalid type definition";
						*outmessage = Tcl_Alloc(strlen(r)+1); 
						strcpy(*outmessage,r);
					}
					SQLFreeHandle(SQL_HANDLE_STMT,newstatement->stmt);
					Tcl_Free((char *) aD); 
					Tcl_Free((char *)newstatement); 
					return NULL;
				}
			}
		}
	}	
	return newstatement;
}

static void UnbindAndFreeColumnBuffers(struct statement *statement) {
	int ib;
	SQLHSTMT stmt = statement->stmt;
	struct ColumnBuffer *buffers = statement->columnBuffers;
	int count = statement->numColumns;
	SQLFreeStmt(stmt,SQL_UNBIND);
	for (ib = 0; ib < count; ib++) {
		if (buffers[ib].valueBuffer != NULL) Tcl_Free((char *)buffers[ib].valueBuffer);
	}
	Tcl_Free((char *)buffers);
	statement->columnBuffers = NULL;
	statement->numColumns = -1;
}
		
static int ExecuteStatement(Tcl_Interp *interp,struct statement *statement,Tcl_Obj *args,char **outmessage) {
	int numArgs, iArg;
	Tcl_Obj *theArg;
    	const char *tname = NULL;
    	int   tnameLen = 0;
	SQLRETURN ret;
	if (args == NULL) numArgs = 0;
	else if (Tcl_ListObjLength(interp,args,&numArgs) != TCL_OK) {
		if (outmessage != NULL) {
			const char *r = Tcl_GetStringResult(interp);
			*outmessage = Tcl_Alloc(strlen(r)+1);
			strcpy(*outmessage,r);
		}
		return FALSE;
	}		
	if (statement->columnBuffers != NULL) {
		UnbindAndFreeColumnBuffers(statement);
	}
	if (statement->statementType != STMT_SQL) {
		if (numArgs > 0) {
			Tcl_ListObjIndex(interp,args,0,&theArg);
			tname = Tcl_GetStringFromObj(theArg,&tnameLen);
		}
	} else {
		int i;
		struct ArgDef *aD = statement->argDefs;
		if (numArgs > statement->numArgDefs)  {
			if (outmessage != NULL) {
				const char *r = "Too many arguments";
				*outmessage = Tcl_Alloc(strlen(r)+1);
				strcpy(*outmessage,r);
			}
			return FALSE;
		}
		for (i = 0; i < numArgs; i++) {
			Tcl_ListObjIndex(interp,args,i,&theArg);
			if (theArg == NULL) {
				tname = NULL;
				tnameLen = 0;
			}  else {
				tname = Tcl_GetStringFromObj(theArg,&tnameLen);
			}
			if (tnameLen == 0) {
				aD[i].valuePointer = SQL_NULL_DATA;
			} else {
				aD[i].valuePointer = tnameLen;
				if (aD[i].columnDef == 0) aD[i].columnDef = tnameLen;
			}
			ret = SQLBindParameter(statement->stmt,i+1,SQL_PARAM_INPUT,
						aD[i].sourceType,aD[i].sqlType,
						aD[i].columnDef,aD[i].scale,
						(SQLPOINTER)tname,tnameLen+1,
						&aD[i].valuePointer);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
				return FALSE;
			}
		}
	}
	ret = SQLFreeStmt(statement->stmt, SQL_CLOSE);
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
		return FALSE;
	}
	switch (statement->statementType) {
	    case STMT_TABLES: /* SQLTables() */
		while ((ret = SQLTables(statement->stmt,NULL,0,NULL,0,tname,tnameLen,"TABLES",SQL_NTS))
			== SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		break;
	    case STMT_COLUMNS: /* SQLColumns() */
		while ((ret = SQLColumns(statement->stmt,NULL,0,NULL,0,tname,tnameLen,NULL,0))
			== SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		break;
	    case STMT_INDEXES: /* SQLStatisics() */
		if (numArgs < 1) {
			if (outmessage != NULL) {
				const char *r = "Missing table name argument!";
				*outmessage = Tcl_Alloc(strlen(r)+1);
				strcpy(*outmessage,r);
			}
			return FALSE;
		}
		while ((ret = SQLStatistics(statement->stmt,NULL,0,NULL,0,tname,tnameLen,SQL_INDEX_ALL,SQL_QUICK))
			== SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		break;
	    case STMT_TYPEINFO: /* SQLTypeInfo() */
		if (numArgs < 1) {
			if (outmessage != NULL) {
				const char *r = "Missing typeid argument!";
				*outmessage = Tcl_Alloc(strlen(r)+1);
				strcpy(*outmessage,r);
			}
			return FALSE;
		}
		SQLSMALLINT datatype = LookUpSqlType(tname);
		while ((ret = SQLGetTypeInfo(statement->stmt,datatype))
			== SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		break;
	    case STMT_PRIMARYKEYS: /* SQLPrimaryKeys() */
		if (numArgs < 1) {
			if (outmessage != NULL) {
				const char *r = "Missing table name argument!";
				*outmessage = Tcl_Alloc(strlen(r)+1);
				strcpy(*outmessage,r);
			}
			return FALSE;
		}
		while ((ret = SQLPrimaryKeys(statement->stmt,NULL,0,NULL,0,tname,tnameLen))
			== SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
	        break;
	    case STMT_SQL: /* SQLExecute() */
	        while ((ret = SQLExecute(statement->stmt)) == SQL_STILL_EXECUTING) EventWait(1);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("ExecuteStatement",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		break;
	}
	return TRUE;
}

static int AllocAndBindColumnBuffers(struct statement *statement,char **outmessage) {
	SQLRETURN ret;
	struct ColumnBuffer *buffers;
	int i;
	SQLHSTMT stmt = statement->stmt;
	int count = statement->numColumns;
	int unbound_cols = FALSE;
	buffers = statement->columnBuffers = (struct ColumnBuffer *)
			Tcl_Alloc(count*sizeof(struct ColumnBuffer));
	memset(buffers,0,count*sizeof(struct ColumnBuffer));
	for (i = 0; i < count; i++)  {
		while ((ret = SQLColAttribute(stmt,i+1,SQL_COLUMN_DISPLAY_SIZE,
					NULL, 0, NULL, &buffers[i].bufferSize))
			== SQL_STILL_EXECUTING) Tcl_Sleep(0);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("AllocAndBindColumnBuffers",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		while ((ret = SQLColAttribute(stmt,i+1,SQL_COLUMN_TYPE,
					NULL, 0, NULL, &buffers[i].sqlType))
			== SQL_STILL_EXECUTING) Tcl_Sleep(0);
		if (!SQL_SUCCEEDED(ret)) {
			extract_error("AllocAndBindColumnBuffers",statement->stmt,SQL_HANDLE_STMT,outmessage);
			return FALSE;
		}
		buffers[i].targetType = MapSqlType(buffers[i].sqlType);
		if (buffers[i].bufferSize == SQL_NO_TOTAL ||
		    buffers[i].sqlType == SQL_LONGVARBINARY ||
		    buffers[i].sqlType == SQL_LONGVARCHAR ||
		    buffers[i].bufferSize > MAX_BLOB_COLUMN_LENGTH) {
			buffers[i].valueBuffer = NULL;
			buffers[i].bufferSize = 0;
			buffers[i].boundP = FALSE;
			if (!CheckDriverExtension(statement->parent,SQL_GD_ANY_COLUMN)) {
				unbound_cols = TRUE;
			}
		} else {
			buffers[i].valueBuffer = Tcl_Alloc((buffers[i].bufferSize+1)*sizeof(char));
			memset(buffers[i].valueBuffer,0,(buffers[i].bufferSize+1)*sizeof(char));
			if (!unbound_cols) {
				while ((ret = SQLBindCol(stmt,i+1,buffers[i].targetType,
							 buffers[i].valueBuffer,buffers[i].bufferSize+1,
							 &(buffers[i].indicator)))== SQL_STILL_EXECUTING) Tcl_Sleep(0);
				if (!SQL_SUCCEEDED(ret)) {
					extract_error("AllocAndBindColumnBuffers",statement->stmt,SQL_HANDLE_STMT,outmessage);
					return FALSE;
				}
				buffers[i].boundP = TRUE;
			} else {
				buffers[i].boundP = FALSE;
			}
		}
	}
	return TRUE;
}

static int FetchResultRow(struct statement *statement,char **outmessage) {
	SQLRETURN ret;
	int  i;
	struct ColumnBuffer *buffers;
	/* Things to do: */
	/* 1) Allocate & bind columnBuffers if NULL; */
	if (NumResultCols(statement,outmessage) <= 0) return FALSE;
#ifdef DEBUG
	fprintf(stderr,"*** FetchResultRow: statement->numColumns = %d\n",statement->numColumns);
#endif
	if (statement->columnBuffers == NULL && statement->numColumns > 0)
		if (!AllocAndBindColumnBuffers(statement,outmessage)) return FALSE;
	/* 2) call SQLFetch() */
	while ((ret = SQLFetch(statement->stmt)) == SQL_STILL_EXECUTING) EventWait(0);
	/* 3) If SQL_NO_DATA, unbind and free columnBuffers, return false. */
#ifdef DEBUG
	fprintf(stderr,"*** FetchResultRow (SQLFetch): ret = %d\n",ret);
#endif
	if (ret == SQL_NO_DATA) return FALSE;
	/* 4) Get data and return true. */
	if (!SQL_SUCCEEDED(ret)) {
		extract_error("FetchResultRow (SQLFetch)",statement->stmt,SQL_HANDLE_STMT,outmessage);
		return FALSE;
	}
	buffers = statement->columnBuffers;
	SQLHSTMT stmt = statement->stmt;
	for (i = 0; i < statement->numColumns; i++) {
		if (buffers[i].valueBuffer == NULL) {
			char dummy;
			while ((ret = SQLGetData(stmt,i+1,buffers[i].targetType,
						 &dummy,
						 buffers[i].targetType == SQL_C_CHAR ? 1 : 0,
						 &(buffers[i].indicator))) == SQL_STILL_EXECUTING) Tcl_Sleep(0);
#ifdef DEBUG
			fprintf(stderr,"*** FetchResultRow (SQLGetData [no buffer]): ret = %d\n",ret);
#endif
			if (!SQL_SUCCEEDED(ret)  && ret != SQL_NO_DATA) {
				extract_error("FetchResultRow (SQLGetData [no buffer])",statement->stmt,SQL_HANDLE_STMT,outmessage);
				return FALSE;
			}
		} else if (!buffers[i].boundP) {
			while ((ret = SQLGetData(stmt,i+1,buffers[i].targetType,
						 buffers[i].valueBuffer,
						 buffers[i].bufferSize,
						 &(buffers[i].indicator))) == SQL_STILL_EXECUTING) Tcl_Sleep(0);
#ifdef DEBUG
			fprintf(stderr,"*** FetchResultRow (SQLGetData [not bound]): ret = %d\n",ret);
#endif
			if (!SQL_SUCCEEDED(ret) && ret != SQL_NO_DATA) {
				extract_error("FetchResultRow (SQLGetData [not bound])",statement->stmt,SQL_HANDLE_STMT,outmessage);
				return FALSE;
			}
		}
		if (buffers[i].indicator == SQL_NULL_DATA  || ret == SQL_NO_DATA) {
			buffers[i].tclValue = Tcl_NewObj();
		} else {
			if (buffers[i].valueBuffer != NULL) {
				if (ret == SQL_NO_DATA) buffers[i].tclValue = Tcl_NewObj();
				else buffers[i].tclValue =
				    Tcl_NewStringObj(buffers[i].valueBuffer,
						     min(buffers[i].indicator,
							 buffers[i].bufferSize));
			} else if (buffers[i].indicator == SQL_NO_TOTAL) {
				buffers[i].tclValue = Tcl_NewObj();
#define BUFSIZE 65536
				char *buf = Tcl_Alloc(BUFSIZE+1);
				while (buffers[i].indicator == SQL_NO_TOTAL ||
				       buffers[i].indicator > BUFSIZE) {
				    while ((ret = SQLGetData(stmt,i+1,
							     buffers[i].targetType,
							     buf,
							     BUFSIZE+(buffers[i].targetType == SQL_C_CHAR ? 1 : 0),
							     &(buffers[i].indicator))) == SQL_STILL_EXECUTING) Tcl_Sleep(0);
#ifdef DEBUG
				    fprintf(stderr,"*** FetchResultRow (SQLGetData [chunk]): ret = %d\n",ret);
#endif
				    if (!SQL_SUCCEEDED(ret) &&  ret  != SQL_NO_DATA) {
					extract_error("FetchResultRow (SQLGetData [chunk])",statement->stmt,SQL_HANDLE_STMT,outmessage);
					return FALSE;
				    }
				    if (ret != SQL_NO_DATA) 
					Tcl_AppendToObj(buffers[i].tclValue,
						    buf,
						    buffers[i].indicator == SQL_NO_TOTAL ? BUFSIZE : buffers[i].indicator);
				}
				Tcl_Free(buf);
#undef BUFSIZE
			} else {
				char *buf = Tcl_Alloc(buffers[i].indicator+1);
				while ((ret = SQLGetData(stmt,i+1,
							     buffers[i].targetType,
							     buf,
							     buffers[i].indicator+(buffers[i].targetType == SQL_C_CHAR ? 1 : 0),
							     &(buffers[i].indicator))) == SQL_STILL_EXECUTING) Tcl_Sleep(0);
#ifdef DEBUG
				fprintf(stderr,"*** FetchResultRow (SQLGetData [no buffer, but with size]): ret = %d\n",ret);
#endif
				if (!SQL_SUCCEEDED(ret) && ret != SQL_NO_DATA) {
					extract_error("FetchResultRow (SQLGetData [no buffer, but with size]",statement->stmt,SQL_HANDLE_STMT,outmessage);
					return FALSE;
				}
				if (ret == SQL_NO_DATA) buffers[i].tclValue = Tcl_NewObj();
				else  buffers[i].tclValue = Tcl_NewStringObj(buf,buffers[i].indicator);
				Tcl_Free(buf);
			}
		}
	}
	return TRUE;
}

int GetColumnLabels(Tcl_Interp *interp,struct statement *statement,Tcl_Obj **labels)
{
	SQLRETURN ret;
	char nameBuff[256];
	SQLSMALLINT col, count;
	char *message = NULL;
	Tcl_Obj *labelList = Tcl_NewObj();
	count = NumResultCols(statement,&message);
	if (count < 0) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}
	for (col = 1; col <= count; col++) {
		ret = SQLColAttribute(statement->stmt,col,SQL_COLUMN_LABEL,
					nameBuff,256,NULL,NULL);
#ifdef DEBUG
		fprintf(stderr,"*** GetColumnLabels: ret = %d, col = %d, nameBuff = '%s'\n",ret,col,nameBuff);
#endif
		if (!SQL_SUCCEEDED(ret)) {
			extract_error_tointerp("GetColumnLabels",statement->stmt,SQL_HANDLE_STMT,interp);
			return TCL_ERROR;
		}
		if (Tcl_ListObjAppendElement(interp,labelList,Tcl_NewStringObj(nameBuff,-1)) != TCL_OK) return TCL_ERROR;
	}
	*labels = labelList;
	return TCL_OK;
}



static void DestroyStatement(struct statement *statement) {
	UnbindAndFreeColumnBuffers(statement);
	if (statement->argDefs != NULL) Tcl_Free((char *)statement->argDefs);
	SQLFreeStmt(statement->stmt,SQL_CLOSE);
	SQLFreeHandle(SQL_HANDLE_STMT,statement->stmt);
	Tcl_Free((char *)statement);
}

static int RunStatement(Tcl_Interp *interp,struct statement *statement,
			Tcl_Obj *args) {
	SQLRETURN ret;
	char *message = NULL;
	int count, i, rows;
	struct ColumnBuffer *buffers;
	Tcl_Obj *results, *rowlist;
	if  (!ExecuteStatement(interp,statement,args,&message)) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}
	/* Fetch Results */
	results = NULL;
	while (1) {
		rows = 0;
		while (FetchResultRow(statement,&message)) {
			/* Acclumulate results list-of-lists */
			rowlist = Tcl_NewObj();
			count = statement->numColumns;
			buffers = statement->columnBuffers;
			for (i = 0; i < count; i++) {
				if (Tcl_ListObjAppendElement(interp,rowlist,buffers[i].tclValue) != TCL_OK) return TCL_ERROR;
			}
			if (results == NULL) results = Tcl_NewObj();
			if (Tcl_ListObjAppendElement(interp,results,rowlist) != TCL_OK) return TCL_ERROR;
			rows++;
		}
		if (rows == 0) {
		    SQLRETURN ret;
		    SQLINTEGER rcount = -2;
		    ret = SQLRowCount(statement->stmt,&rcount);
		    if (SQL_SUCCEEDED(ret)) {
			if (rcount < 0) {
			    if (results == NULL) results = Tcl_NewStringObj("OK",2);
			    else if (Tcl_ListObjAppendElement(interp,results,Tcl_NewStringObj("OK",2)) != TCL_OK) return TCL_ERROR;
			} else {
			    Tcl_Obj *answer = Tcl_NewLongObj(rcount);
			    Tcl_AppendToObj(answer," ROWS",-1);
			    if (results == NULL) results = answer;
			    else if (Tcl_ListObjAppendElement(interp,results,answer) != TCL_OK) return TCL_ERROR;
		        }
		    } else {
		    	if (results == NULL) results = Tcl_NewStringObj("OK",2);
		    	else if (Tcl_ListObjAppendElement(interp,results,Tcl_NewStringObj("OK",2)) != TCL_OK) return TCL_ERROR;
		    }
		}
		UnbindAndFreeColumnBuffers(statement);
		if (message !=  NULL) {
			Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
			Tcl_Free(message);
			return TCL_ERROR;
		}			
		if  (!statement->multiSetsP) break;
		while ((ret = SQLMoreResults(statement->stmt))
		       == SQL_STILL_EXECUTING) EventWait(0);
		if (ret == SQL_ERROR) {
			extract_error_tointerp("RunStatement",statement->stmt,SQL_HANDLE_STMT,interp);
			return TCL_ERROR;
		} else if (ret == SQL_NO_DATA) break;
	}
	Tcl_SetObjResult(interp,results);
	return TCL_OK;
}
static int EvalStatement(Tcl_Interp *interp,Tcl_Obj *proc,
			 struct statement *statement,Tcl_Obj *args) {
	SQLRETURN ret;
	int tclRet;
	char *message = NULL;
	Tcl_Obj *commandlist;
	int count, i;
	struct ColumnBuffer *buffers;
	if  (!ExecuteStatement(interp,statement,args,&message)) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}
	/* Fetch Results */
	while (FetchResultRow(statement,&message)) {
		/* Call proc on each row */
		commandlist = Tcl_DuplicateObj(proc);
		count = statement->numColumns;
		buffers = statement->columnBuffers;
		for (i = 0; i < count; i++) {
			if (Tcl_ListObjAppendElement(interp,commandlist,buffers[i].tclValue) != TCL_OK) return TCL_ERROR;
		}
		tclRet = Tcl_EvalObjEx(interp,commandlist,TCL_EVAL_GLOBAL);
		if (tclRet != TCL_OK) return tclRet;
	}
	if (message !=  NULL) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}			
	return TCL_OK;
}
static int ReadStatement(Tcl_Interp *interp,Tcl_Obj *arrayspec,
			 struct statement *statement,Tcl_Obj *args) {
	SQLRETURN ret;
	char *message = NULL;
	Tcl_Obj *arrayName, *arrayIndex, *columnLabels, *value;
	int arraySpecLen;
	int count, i, createSubIndex;
	struct ColumnBuffer *buffers;
	if (Tcl_ListObjLength(interp,arrayspec,&arraySpecLen) != TCL_OK) return TCL_ERROR;
	if  (!ExecuteStatement(interp,statement,args,&message)) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}
	count = NumResultCols(statement,&message);
	if (count < 0) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	} else if (count == 0) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj("No Output Colums!",-1));
		return TCL_ERROR;
	} else if (count == 1) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj("Only one output column!",-1));
		return TCL_ERROR;
	}
	count--;	/* Gobble first column (used for array index) */
	if (count == arraySpecLen) {
		createSubIndex = FALSE;
	} else if (arraySpecLen == 1) {
		createSubIndex = TRUE;
		if (GetColumnLabels(interp,statement,&columnLabels) != TCL_OK) return TCL_ERROR;
		arrayName = arrayspec;
	} else {
		Tcl_Obj *result = Tcl_NewStringObj("Invalid array specification: ",-1);
		Tcl_AppendObjToObj(result,arrayspec);
		Tcl_SetObjResult(interp,result);
		return TCL_ERROR;
	}
	/* Fetch Results */
	while (FetchResultRow(statement,&message)) {
		/* Set array elements */
		buffers = statement->columnBuffers;
		for (i = 0; i < count; i++) {
			arrayIndex = Tcl_DuplicateObj(buffers[0].tclValue);
			value = buffers[i+1].tclValue;
			if (createSubIndex) {
				Tcl_Obj *columnLabel;
				Tcl_AppendToObj(arrayIndex,",",1);
				Tcl_ListObjIndex(interp,columnLabels,i+1,&columnLabel);
				Tcl_AppendObjToObj(arrayIndex,columnLabel);
			} else {
				Tcl_ListObjIndex(interp,arrayspec,i,&arrayName);
			}
			if (Tcl_ObjSetVar2(interp,arrayName,arrayIndex,value,TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
		}
	}
	if (message !=  NULL) {
		Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
		Tcl_Free(message);
		return TCL_ERROR;
	}			
	return TCL_OK;
}


%}

struct statement {
};

#if SWIGVERSIONMINOR == 1 && SWIGVERSIONMAJOR == 1
%addmethods
#else
%extend
#endif
	statement {
	~statement() {
		DestroyStatement(self);
	}
	Tcl_Result cget(Tcl_Interp *interp,const char *what) {
		SQLRETURN ret;
		SQLUINTEGER tempint;
		if (strncasecmp(what,"-attr_async_enable",strlen(what)) == 0) {
		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_ASYNC_ENABLE,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_async_enable",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
		    	case SQL_ASYNC_ENABLE_OFF:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
			    return TCL_OK;
			case SQL_ASYNC_ENABLE_ON:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_concurrency",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_CONCURRENCY,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_concurrency",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
			case SQL_CONCUR_READ_ONLY:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("read_only",-1));
			    return TCL_OK;
			case SQL_CONCUR_LOCK:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("lock",-1));
			    return TCL_OK;
			case SQL_CONCUR_ROWVER:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("rowver",-1));
			    return TCL_OK;
			case SQL_CONCUR_VALUES:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("values",-1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_cursor_scrollable",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_CURSOR_SCROLLABLE,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_cursor_scrollable",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
		    	case SQL_NONSCROLLABLE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("nonscrollable",-1));
			    return TCL_OK;
			case SQL_SCROLLABLE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("scrollable",-1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_cursor_sensitivity",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_CURSOR_SENSITIVITY,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_cursor_sensitivity",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
		    	case SQL_UNSPECIFIED:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("unspecified",-1));
			    return TCL_OK;
			case SQL_INSENSITIVE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("insensitive",-1));
			    return TCL_OK;
			case SQL_SENSITIVE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("sensitive",-1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_cursor_type",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_CURSOR_TYPE,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_cursor_type",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
		    	case SQL_CURSOR_FORWARD_ONLY:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("forward_only",-1));
			    return TCL_OK;
			case SQL_CURSOR_STATIC:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("static",-1));
			    return TCL_OK;
			case SQL_CURSOR_KEYSET_DRIVEN:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("keyset_driven",-1));
			    return TCL_OK;
			case SQL_CURSOR_DYNAMIC:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("dynamic",-1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_enable_auto_ipd",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_ENABLE_AUTO_IPD,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_enable_auto_ipd",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
			case SQL_FALSE:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
			    return TCL_OK;
			case SQL_TRUE:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_keyset_size",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_KEYSET_SIZE,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_keyset_size",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    Tcl_SetObjResult(interp,Tcl_NewLongObj(tempint));
		    return TCL_OK;
		} else if (strncasecmp(what,"-attr_max_length",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_MAX_LENGTH,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_max_length",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    Tcl_SetObjResult(interp,Tcl_NewLongObj(tempint));
		    return TCL_OK;
		} else if (strncasecmp(what,"-attr_max_rows",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_MAX_ROWS,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_max_rows",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    Tcl_SetObjResult(interp,Tcl_NewLongObj(tempint));
		    return TCL_OK;
		} else if (strncasecmp(what,"-attr_metadata_id",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_METADATA_ID,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_metadata_id",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
			case SQL_FALSE:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
			    return TCL_OK;
			case SQL_TRUE:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_noscan",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_NOSCAN,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_noscan",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
			case SQL_NOSCAN_OFF:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
			    return TCL_OK;
			case SQL_NOSCAN_ON:
			    Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_query_timeout",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_QUERY_TIMEOUT,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_query_timeout",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    Tcl_SetObjResult(interp,Tcl_NewLongObj(tempint));
		    return TCL_OK;
		} else if (strncasecmp(what,"-attr_row_number",strlen(what)) == 0) {
		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_ROW_NUMBER,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_row_number",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    Tcl_SetObjResult(interp,Tcl_NewLongObj(tempint));
		    return TCL_OK;
		} else if (strncasecmp(what,"-attr_simulate_cursor",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_SIMULATE_CURSOR,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_simulate_cursor",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
			case SQL_SC_NON_UNIQUE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("non_unique",-1));
			    return TCL_OK;
			case SQL_SC_TRY_UNIQUE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("try_unique",-1));
			    return TCL_OK;
			case SQL_SC_UNIQUE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("unique",-1));
			    return TCL_OK;
			default:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("unknown",-1));
			    return TCL_OK;
		    }
		} else if (strncasecmp(what,"-attr_use_bookmarks",strlen(what)) == 0) {
 		    ret = SQLGetStmtAttr(self->stmt,SQL_ATTR_USE_BOOKMARKS,
		    			 (SQLPOINTER)&tempint,SQL_IS_UINTEGER,
					 NULL);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("cget -attr_use_bookmarks",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		    switch (tempint) {
		    	case SQL_UB_OFF:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("off",-1));
			    return TCL_OK;
		    	case SQL_UB_VARIABLE:
			    Tcl_SetObjResult(interp,Tcl_NewStringObj("variable",-1));
			    return TCL_OK;
		    }
		/* Copied from SWIG Generated code */
#if SWIGVERSIONMINOR == 1 && SWIGVERSIONMAJOR == 1
			SWIG_MakePtr(interp->result,(void *) self, "_statement_p");
			return TCL_OK;
#else
		} else if (strcmp(what,"-this") == 0) {
			Tcl_SetObjResult(interp, SWIG_NewInstanceObj((void *) self,SWIGTYPE_p_statement,0));
			return TCL_OK;
		} else if (strcmp(what,"-thisown") == 0) {
			if (SWIG_Thisown(self)) 
				Tcl_SetObjResult(interp, Tcl_NewBooleanObj(1));
			else	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(0));
			return TCL_OK;
#endif
		} else {
			Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,what,-1);
			Tcl_AppendToObj(result,"\", should be one of: ",-1);
			Tcl_AppendToObj(result,"-attr_async_enable ",-1);
			Tcl_AppendToObj(result,"-attr_concurrency ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_scrollable ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_sensitivity ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_type ",-1);
			Tcl_AppendToObj(result,"-attr_enable_auto_ipd ",-1);
			Tcl_AppendToObj(result,"-attr_keyset_size ",-1);
			Tcl_AppendToObj(result,"-attr_max_length ",-1);
			Tcl_AppendToObj(result,"-attr_max_rows ",-1);
			Tcl_AppendToObj(result,"-attr_metadata_id ",-1);
			Tcl_AppendToObj(result,"-attr_noscan ",-1);
			Tcl_AppendToObj(result,"-attr_query_timeout ",-1);
			Tcl_AppendToObj(result,"-attr_row_number ",-1);
			Tcl_AppendToObj(result,"-attr_simulate_cursor ",-1);
			Tcl_AppendToObj(result,"-attr_use_bookmarks ",-1);
			Tcl_AppendToObj(result,"-this ",-1);
#if SWIGVERSIONMINOR == 1 && SWIGVERSIONMAJOR == 1
#else
			Tcl_AppendToObj(result,"-thisown ",-1);
#endif
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	Tcl_Result configure(Tcl_Interp *interp,const char *what,Tcl_Obj *value) {
		SQLRETURN ret;
		SQLUINTEGER tempint;
 		if (strncasecmp(what,"-attr_async_enable",strlen(what)) == 0) {
		    int btemp;
		    if (Tcl_GetBooleanFromObj(interp,value,&btemp) != TCL_OK) return TCL_ERROR;
		    if (btemp)	tempint = SQL_ASYNC_ENABLE_ON;
		    else	tempint = SQL_ASYNC_ENABLE_OFF;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_ASYNC_ENABLE,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_async_enable",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_concurrency",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"read_only",strlen(vs)) == 0)
		    	tempint = SQL_CONCUR_READ_ONLY;
		    else if (strncasecmp(vs,"lock",strlen(vs)) == 0)
		    	tempint = SQL_CONCUR_LOCK;
		    else if (strncasecmp(vs,"rowver",strlen(vs)) == 0)
		    	tempint = SQL_CONCUR_ROWVER;
		    else if (strncasecmp(vs,"values",strlen(vs)) == 0)
		    	tempint = SQL_CONCUR_VALUES;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_CONCURRENCY,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_concurrency",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_cursor_scrollable",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"nonscrollable",strlen(vs)) == 0)
		    	tempint = SQL_NONSCROLLABLE;
		    else if (strncasecmp(vs,"scrollable",strlen(vs)) == 0)
		    	tempint = SQL_SCROLLABLE;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_CURSOR_SCROLLABLE,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_cursor_scrollable",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_cursor_sensitivity",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"unspecified",strlen(vs)) == 0)
			tempint = SQL_UNSPECIFIED;
		    else if (strncasecmp(vs,"insensitive",strlen(vs)) == 0)
			tempint = SQL_INSENSITIVE;
		    else if (strncasecmp(vs,"sensitive",strlen(vs)) == 0)
			tempint = SQL_SENSITIVE;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_CURSOR_SENSITIVITY,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_cursor_sensitivity",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_cursor_type",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"forward_only",strlen(vs)) == 0)
		    	tempint = SQL_CURSOR_FORWARD_ONLY;
		    else if (strncasecmp(vs,"static",strlen(vs)) == 0)
		    	tempint = SQL_CURSOR_STATIC;
		    else if (strncasecmp(vs,"keyset_driven",strlen(vs)) == 0)
		    	tempint = SQL_CURSOR_KEYSET_DRIVEN;
		    else if (strncasecmp(vs,"dynamic",strlen(vs)) == 0)
		    	tempint = SQL_CURSOR_DYNAMIC;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_CURSOR_TYPE,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_cursor_type",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_enable_auto_ipd",strlen(what)) == 0) {
		    int btemp;
		    if (Tcl_GetBooleanFromObj(interp,value,&btemp) != TCL_OK) return TCL_ERROR;
		    if (btemp)	tempint = SQL_TRUE;
		    else	tempint = SQL_FALSE;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_ENABLE_AUTO_IPD,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_enable_auto_ipd",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_keyset_size",strlen(what)) == 0) {
		    if (Tcl_GetLongFromObj(interp,value,&tempint) != TCL_OK) return TCL_ERROR;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_KEYSET_SIZE,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_keyset_size",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_max_length",strlen(what)) == 0) {
		    if (Tcl_GetLongFromObj(interp,value,&tempint) != TCL_OK) return TCL_ERROR;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_MAX_LENGTH,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_max_length",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_max_rows",strlen(what)) == 0) {
		    if (Tcl_GetLongFromObj(interp,value,&tempint) != TCL_OK) return TCL_ERROR;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_MAX_ROWS,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_max_rows",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_metadata_id",strlen(what)) == 0) {
		    int btemp;
		    if (Tcl_GetBooleanFromObj(interp,value,&btemp) != TCL_OK) return TCL_ERROR;
		    if (btemp)	tempint = SQL_TRUE;
		    else	tempint = SQL_FALSE;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_METADATA_ID,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_metadata_id",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_noscan",strlen(what)) == 0) {
		    int btemp;
		    if (Tcl_GetBooleanFromObj(interp,value,&btemp) != TCL_OK) return TCL_ERROR;
		    if (btemp)	tempint = SQL_NOSCAN_ON;
		    else	tempint = SQL_NOSCAN_OFF;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_NOSCAN,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_noscan",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_query_timeout",strlen(what)) == 0) {
		    if (Tcl_GetLongFromObj(interp,value,&tempint) != TCL_OK) return TCL_ERROR;
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_QUERY_TIMEOUT,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_query_timeout",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_simulate_cursor",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"non_unique",strlen(vs)) == 0)
		    	tempint = SQL_SC_NON_UNIQUE;
		    else if (strncasecmp(vs,"try_unique",strlen(vs)) == 0)
		    	tempint = SQL_SC_TRY_UNIQUE;
		    else if (strncasecmp(vs,"unique",strlen(vs)) == 0)
		    	tempint = SQL_SC_UNIQUE;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_SIMULATE_CURSOR,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_simulate_cursor",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else if (strncasecmp(what,"-attr_use_bookmarks",strlen(what)) == 0) {
		    const char *vs = Tcl_GetString(value);
		    if (strncasecmp(vs,"off",strlen(vs)) == 0)
		    	tempint = SQL_UB_OFF;
		    else if (strncasecmp(vs,"variable",strlen(vs)) == 0)
		    	tempint = SQL_UB_VARIABLE;
		    else if (strncasecmp(vs,"fixed",strlen(vs)) == 0)
		    	tempint = SQL_UB_VARIABLE;
		    else {
		    	Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,vs,-1);
			Tcl_AppendToObj(result,"\"",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		    }
		    ret = SQLSetStmtAttr(self->stmt,SQL_ATTR_USE_BOOKMARKS,
		    			 (SQLPOINTER)tempint,SQL_IS_UINTEGER);
		    if (!SQL_SUCCEEDED(ret)) {
		    	extract_error_tointerp("configure -attr_use_bookmarks",self->stmt,SQL_HANDLE_STMT,interp);
		    	return TCL_ERROR;
		    }
		} else {
			Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,what,-1);
			Tcl_AppendToObj(result,"\", should be one of: ",-1);
			Tcl_AppendToObj(result,"-attr_async_enable ",-1);
			Tcl_AppendToObj(result,"-attr_concurrency ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_scrollable ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_sensitivity ",-1);
			Tcl_AppendToObj(result,"-attr_cursor_type ",-1);
			Tcl_AppendToObj(result,"-attr_enable_auto_ipd ",-1);
			Tcl_AppendToObj(result,"-attr_keyset_size ",-1);
			Tcl_AppendToObj(result,"-attr_max_length ",-1);
			Tcl_AppendToObj(result,"-attr_max_rows ",-1);
			Tcl_AppendToObj(result,"-attr_metadata_id ",-1);
			Tcl_AppendToObj(result,"-attr_noscan ",-1);
			Tcl_AppendToObj(result,"-attr_query_timeout ",-1);
			Tcl_AppendToObj(result,"-attr_simulate_cursor ",-1);
			Tcl_AppendToObj(result,"-attr_use_bookmarks ",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	Tcl_Result run(Tcl_Interp *interp,Tcl_Obj *args=NULL) {
		return RunStatement(interp,self,args);
	}
	Tcl_Result eval(Tcl_Interp *interp,Tcl_Obj *proc,Tcl_Obj *args=NULL) {
		return EvalStatement(interp,proc,self,args);
	}
	Tcl_Result read(Tcl_Interp *interp,Tcl_Obj *arrayspec,Tcl_Obj *args=NULL) {
		return ReadStatement(interp,arrayspec,self,args);
	}
	Tcl_Result execute(Tcl_Interp *interp,Tcl_Obj *args=NULL) {
		char *message = NULL;
		if  (!ExecuteStatement(interp,self,args,&message)) {
			Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
			Tcl_Free(message);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	Tcl_Result rowcount(Tcl_Interp *interp) {
		SQLINTEGER rcount;
		SQLRETURN ret;
		ret = SQLRowCount(self->stmt,&rcount);
		if (!SQL_SUCCEEDED(ret)) {
		   	extract_error_tointerp("rowcount",self->stmt,SQL_HANDLE_STMT,interp);
		   	return TCL_ERROR;
		}
		Tcl_SetObjResult(interp,Tcl_NewLongObj(rcount));
		return TCL_OK;
	}
	Tcl_Result numresultcols(Tcl_Interp *interp) {
		char *message = NULL;
		int cols = NumResultCols(self,&message);
		if (cols < 0) {
			Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
			Tcl_Free(message);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	Tcl_Result columncget(Tcl_Interp *interp,int columnNumber,const char *what) {
		char *message = NULL;
		SQLRETURN ret;
		SQLINTEGER numResult;
		SQLSMALLINT rlen;
		char charResult[256];
		int cols = NumResultCols(self,&message);
		if (cols < 0) {
			Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
			Tcl_Free(message);
			return TCL_ERROR;
		}
		if (columnNumber < 0 || columnNumber > (cols+1)) {
			Tcl_Obj *err = Tcl_NewStringObj("Column number out of range: ",-1);
			Tcl_AppendObjToObj(err,Tcl_NewIntObj(columnNumber));
			return TCL_ERROR;
		}
		if (strncasecmp(what,"-auto_unique_value",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_AUTO_UNIQUE_VALUE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_TRUE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
				return TCL_OK;
			    case SQL_FALSE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
				return TCL_OK;
			}
		} else if (strncasecmp(what,"-base_column_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_BASE_COLUMN_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-base_table_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_BASE_TABLE_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-case_sensitive",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_CASE_SENSITIVE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_TRUE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
				return TCL_OK;
			    case SQL_FALSE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
				return TCL_OK;
			}
		} else if (strncasecmp(what,"-catalog_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_CATALOG_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-display_size",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_DISPLAY_SIZE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
		} else if (strncasecmp(what,"-fixed_prec_scale",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_FIXED_PREC_SCALE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_TRUE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
				return TCL_OK;
			    case SQL_FALSE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
				return TCL_OK;
			}
		} else if (strncasecmp(what,"-label",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_LABEL,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-length",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_LENGTH,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
		} else if (strncasecmp(what,"-literal_prefix",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_LITERAL_PREFIX,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-literal_suffix",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_LITERAL_SUFFIX,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-local_type_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_LOCAL_TYPE_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-nullable",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_NULLABLE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_NO_NULLS:
			    	Tcl_SetObjResult(interp,Tcl_NewStringObj("no_nulls",-1));
			    	return TCL_OK;
			    case SQL_NULLABLE_UNKNOWN:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("nullable_unknown",-1));
			    	return TCL_OK;
			}
#ifdef SQL_DESC_NUM_PREX_RADIX
		} else if (strncasecmp(what,"-num_prex_radix",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_NUM_PREX_RADIX,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
#endif
		} else if (strncasecmp(what,"-octet_length",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_OCTET_LENGTH,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
		} else if (strncasecmp(what,"-precision",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_PRECISION,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
		} else if (strncasecmp(what,"-scale",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_SCALE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewLongObj(numResult));
			return TCL_OK;
		} else if (strncasecmp(what,"-schema_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_SCHEMA_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-searchable",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_SEARCHABLE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_PRED_NONE:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("none",-1));
				return TCL_OK;
			    case SQL_PRED_CHAR:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("char",-1));
				return TCL_OK;
			    case SQL_PRED_BASIC:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("basic",-1));
				return TCL_OK;
			    case SQL_PRED_SEARCHABLE:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("searchable",-1));
				return TCL_OK;
			}
		} else if (strncasecmp(what,"-table_name",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_TABLE_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-type",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_TYPE_NAME,
						charResult,256,&rlen,NULL);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			Tcl_SetObjResult(interp,Tcl_NewStringObj(charResult,rlen));
			return TCL_OK;
		} else if (strncasecmp(what,"-unnamed",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_UNNAMED,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_NAMED:
			    	Tcl_SetObjResult(interp,Tcl_NewStringObj("named",-1));
			    	return TCL_OK;
			    case SQL_UNNAMED:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("unnamed",-1));
			    	return TCL_OK;
			}
		} else if (strncasecmp(what,"-unsigned",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_UNSIGNED,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_TRUE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
				return TCL_OK;
			    case SQL_FALSE:
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
				return TCL_OK;
			}
		} else if (strncasecmp(what,"-updatable",strlen(what)) == 0) {
			ret = SQLColAttribute(self->stmt,columnNumber,
						SQL_DESC_UPDATABLE,
						NULL, 0, NULL, &numResult);
			if (!SQL_SUCCEEDED(ret)) {
				extract_error_tointerp("columncget",self->stmt,SQL_HANDLE_STMT,interp);
				return TCL_ERROR;
			}
			switch (numResult) {
			    case SQL_ATTR_READONLY:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("readonly",-1));
				return TCL_OK;
			    case SQL_ATTR_WRITE:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("write",-1));
				return TCL_OK;
			    case SQL_ATTR_READWRITE_UNKNOWN:
				Tcl_SetObjResult(interp,Tcl_NewStringObj("readwrite_unknown",-1));
				return TCL_OK;
			}
		} else {
			Tcl_Obj  *result = Tcl_NewStringObj("Unknown option: \"",-1);
			Tcl_AppendToObj(result,what,-1);
			Tcl_AppendToObj(result,"\", should be one of: ",-1);
			Tcl_AppendToObj(result,"-auto_unique_value ",-1);
			Tcl_AppendToObj(result,"-base_column_name ",-1);
			Tcl_AppendToObj(result,"-base_table_name ",-1);
			Tcl_AppendToObj(result,"-case_sensitive ",-1);
			Tcl_AppendToObj(result,"-catalog_name ",-1);
			Tcl_AppendToObj(result,"-display_size ",-1);
			Tcl_AppendToObj(result,"-fixed_prec_scale ",-1);
			Tcl_AppendToObj(result,"-label ",-1);
			Tcl_AppendToObj(result,"-length ",-1);
			Tcl_AppendToObj(result,"-literal_prefix ",-1);
			Tcl_AppendToObj(result,"-literal_suffix ",-1);
			Tcl_AppendToObj(result,"-local_type_name ",-1);
			Tcl_AppendToObj(result,"-name ",-1);
			Tcl_AppendToObj(result,"-nullable ",-1);
			Tcl_AppendToObj(result,"-num_prex_radix ",-1);
			Tcl_AppendToObj(result,"-octet_length ",-1);
			Tcl_AppendToObj(result,"-precision ",-1);
			Tcl_AppendToObj(result,"-scale ",-1);
			Tcl_AppendToObj(result,"-schema_name ",-1);
			Tcl_AppendToObj(result,"-searchable ",-1);
			Tcl_AppendToObj(result,"-table_name ",-1);
			Tcl_AppendToObj(result,"-type ",-1);
			Tcl_AppendToObj(result,"-unnamed ",-1);
			Tcl_AppendToObj(result,"-unsigned ",-1);
			Tcl_AppendToObj(result,"-updatable ",-1);
			Tcl_SetObjResult(interp,result);
			return TCL_ERROR;
		}
		/* call SQLColAttribute() */
		return TCL_OK;
	}
	Tcl_Result moreresults(Tcl_Interp *interp) {
		SQLRETURN ret;
		while ((ret = SQLMoreResults(self->stmt)) 
			== SQL_STILL_EXECUTING) EventWait(0);
		if (ret == SQL_ERROR) {
			extract_error_tointerp("moreresults",self->stmt,SQL_HANDLE_STMT,interp);
			return TCL_ERROR;
		}
		UnbindAndFreeColumnBuffers(self);
		if (ret == SQL_NO_DATA) Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
		else Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
		return TCL_OK;			
	}
	Tcl_Result fetch(Tcl_Interp *interp,Tcl_Obj *array=NULL,Tcl_Obj *columns=NULL) {
		char *message = NULL;
		int i;
		int count;
		if (FetchResultRow(self,&message)) {
			count = self->numColumns;
			/* process results */
			/* Either a row list or set array elements */
			struct ColumnBuffer *buffers = self->columnBuffers;
			if (array ==  NULL) {
				Tcl_Obj *rowlist = Tcl_NewObj();
				for (i = 0; i < count; i++) {
					if (Tcl_ListObjAppendElement(interp,rowlist,buffers[i].tclValue) != TCL_OK) return TCL_ERROR;
				}
				Tcl_SetObjResult(interp,rowlist);
			} else {
				Tcl_Obj *colLabels, *arrayIndex;
				if (columns == NULL) {
					if (GetColumnLabels(interp,self,&colLabels) != TCL_OK) return TCL_ERROR;
				} else {
					int cl;
					if (Tcl_ListObjLength(interp,columns,&cl) != TCL_OK) return TCL_ERROR;
					if (cl < count) {
						Tcl_Obj  *err = Tcl_NewStringObj("Not enough labels: ",-1);
						Tcl_AppendObjToObj(err,Tcl_NewIntObj(cl));
						Tcl_AppendToObj(err," < ",-1);
						Tcl_AppendObjToObj(err,Tcl_NewIntObj(count));
						Tcl_SetObjResult(interp,err);
						return TCL_ERROR;
					}
					colLabels = columns;
				}
				for (i = 0; i < count; i++) {
					Tcl_ListObjIndex(interp,colLabels,i,&arrayIndex);
					if (Tcl_ObjSetVar2(interp,array,arrayIndex,buffers[i].tclValue,TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
				}
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(1));
			}
		} else if (message != NULL) {
			/* Error from FetchResultRow */
			Tcl_SetObjResult(interp,Tcl_NewStringObj(message,-1));
			Tcl_Free(message);
			return  TCL_ERROR;
		} else {
			/* return empty list or false */
			if (array ==  NULL) {
				Tcl_SetObjResult(interp,Tcl_NewObj());
			} else {
				Tcl_SetObjResult(interp,Tcl_NewBooleanObj(0));
			}
		}
		return TCL_OK;
	}
	Tcl_Result columnlabels(Tcl_Interp *interp) {
		Tcl_Obj *labels;
		if (GetColumnLabels(interp,self,&labels) != TCL_OK) return TCL_ERROR;
		Tcl_SetObjResult(interp,labels);
		return TCL_OK;
	}
};

